@@ -19,7 +19,7 @@ from websockets.frames import CloseCode, Frame, Opcode
 from websockets.protocol import CLIENT, SERVER, Protocol, State
 
 from ..protocol import RecordingProtocol
-from ..utils import MS
+from ..utils import MS, AssertNoLogsMixin
 from .connection import InterceptingConnection
 from .utils import alist
 
@@ -28,7 +28,7 @@ from .utils import alist
 # All tests run on the client side and the server side to validate this.
 
 
-class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
+class ClientConnectionTests(AssertNoLogsMixin, unittest.IsolatedAsyncioTestCase):
     LOCAL = CLIENT
     REMOTE = SERVER
 
@@ -48,23 +48,6 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         await self.remote_connection.close()
         await self.connection.close()
 
-    if sys.version_info[:2] < (3, 10):  # pragma: no cover
-
-        @contextlib.contextmanager
-        def assertNoLogs(self, logger="websockets", level=logging.ERROR):
-            """
-            No message is logged on the given logger with at least the given level.
-
-            """
-            with self.assertLogs(logger, level) as logs:
-                # We want to test that no log message is emitted
-                # but assertLogs expects at least one log message.
-                logging.getLogger(logger).log(level, "dummy")
-                yield
-
-            level_name = logging.getLevelName(level)
-            self.assertEqual(logs.output, [f"{level_name}:{logger}:dummy"])
-
     # Test helpers built upon RecordingProtocol and InterceptingConnection.
 
     async def assertFrameSent(self, frame):
@@ -190,13 +173,13 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         await self.remote_connection.send(b"\x01\x02\xfe\xff")
         self.assertEqual(await self.connection.recv(), b"\x01\x02\xfe\xff")
 
-    async def test_recv_encoded_text(self):
-        """recv receives an UTF-8 encoded text message."""
+    async def test_recv_text_as_bytes(self):
+        """recv receives a text message as bytes."""
         await self.remote_connection.send("😀")
         self.assertEqual(await self.connection.recv(decode=False), "😀".encode())
 
-    async def test_recv_decoded_binary(self):
-        """recv receives an UTF-8 decoded binary message."""
+    async def test_recv_binary_as_text(self):
+        """recv receives a binary message as a str."""
         await self.remote_connection.send("😀".encode())
         self.assertEqual(await self.connection.recv(decode=True), "😀")
 
@@ -222,6 +205,15 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         with self.assertRaises(ConnectionClosedError):
             await self.connection.recv()
 
+    async def test_recv_non_utf8_text(self):
+        """recv receives a non-UTF-8 text message."""
+        await self.remote_connection.send(b"\x01\x02\xfe\xff", text=True)
+        with self.assertRaises(ConnectionClosedError):
+            await self.connection.recv()
+        await self.assertFrameSent(
+            Frame(Opcode.CLOSE, b"\x03\xefinvalid start byte at position 2")
+        )
+
     async def test_recv_during_recv(self):
         """recv raises ConcurrencyError when called concurrently."""
         recv_task = asyncio.create_task(self.connection.recv())
@@ -304,16 +296,16 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
             [b"\x01\x02\xfe\xff"],
         )
 
-    async def test_recv_streaming_encoded_text(self):
-        """recv_streaming receives an UTF-8 encoded text message."""
+    async def test_recv_streaming_text_as_bytes(self):
+        """recv_streaming receives a text message as bytes."""
         await self.remote_connection.send("😀")
         self.assertEqual(
             await alist(self.connection.recv_streaming(decode=False)),
             ["😀".encode()],
         )
 
-    async def test_recv_streaming_decoded_binary(self):
-        """recv_streaming receives a UTF-8 decoded binary message."""
+    async def test_recv_streaming_binary_as_str(self):
+        """recv_streaming receives a binary message as a str."""
         await self.remote_connection.send("😀".encode())
         self.assertEqual(
             await alist(self.connection.recv_streaming(decode=True)),
@@ -352,6 +344,15 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
             async for _ in self.connection.recv_streaming():
                 self.fail("did not raise")
 
+    async def test_recv_streaming_non_utf8_text(self):
+        """recv_streaming receives a non-UTF-8 text message."""
+        await self.remote_connection.send(b"\x01\x02\xfe\xff", text=True)
+        with self.assertRaises(ConnectionClosedError):
+            await alist(self.connection.recv_streaming())
+        await self.assertFrameSent(
+            Frame(Opcode.CLOSE, b"\x03\xefinvalid start byte at position 2")
+        )
+
     async def test_recv_streaming_during_recv(self):
         """recv_streaming raises ConcurrencyError when called concurrently with recv."""
         recv_task = asyncio.create_task(self.connection.recv())
@@ -438,6 +439,16 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         await self.connection.send(b"\x01\x02\xfe\xff")
         self.assertEqual(await self.remote_connection.recv(), b"\x01\x02\xfe\xff")
 
+    async def test_send_binary_from_str(self):
+        """send sends a binary message from a str."""
+        await self.connection.send("😀", text=False)
+        self.assertEqual(await self.remote_connection.recv(), "😀".encode())
+
+    async def test_send_text_from_bytes(self):
+        """send sends a text message from bytes."""
+        await self.connection.send("😀".encode(), text=True)
+        self.assertEqual(await self.remote_connection.recv(), "😀")
+
     async def test_send_fragmented_text(self):
         """send sends a fragmented text message."""
         await self.connection.send(["😀", "😀"])
@@ -456,6 +467,24 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
             [b"\x01\x02", b"\xfe\xff", b""],
         )
 
+    async def test_send_fragmented_binary_from_str(self):
+        """send sends a fragmented binary message from a str."""
+        await self.connection.send(["😀", "😀"], text=False)
+        # websockets sends an trailing empty fragment. That's an implementation detail.
+        self.assertEqual(
+            await alist(self.remote_connection.recv_streaming()),
+            ["😀".encode(), "😀".encode(), b""],
+        )
+
+    async def test_send_fragmented_text_from_bytes(self):
+        """send sends a fragmented text message from bytes."""
+        await self.connection.send(["😀".encode(), "😀".encode()], text=True)
+        # websockets sends an trailing empty fragment. That's an implementation detail.
+        self.assertEqual(
+            await alist(self.remote_connection.recv_streaming()),
+            ["😀", "😀", ""],
+        )
+
     async def test_send_async_fragmented_text(self):
         """send sends a fragmented text message asynchronously."""
 
@@ -484,6 +513,34 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
             [b"\x01\x02", b"\xfe\xff", b""],
         )
 
+    async def test_send_async_fragmented_binary_from_str(self):
+        """send sends a fragmented binary message from a str asynchronously."""
+
+        async def fragments():
+            yield "😀"
+            yield "😀"
+
+        await self.connection.send(fragments(), text=False)
+        # websockets sends an trailing empty fragment. That's an implementation detail.
+        self.assertEqual(
+            await alist(self.remote_connection.recv_streaming()),
+            ["😀".encode(), "😀".encode(), b""],
+        )
+
+    async def test_send_async_fragmented_text_from_bytes(self):
+        """send sends a fragmented text message from bytes asynchronously."""
+
+        async def fragments():
+            yield "😀".encode()
+            yield "😀".encode()
+
+        await self.connection.send(fragments(), text=True)
+        # websockets sends an trailing empty fragment. That's an implementation detail.
+        self.assertEqual(
+            await alist(self.remote_connection.recv_streaming()),
+            ["😀", "😀", ""],
+        )
+
     async def test_send_connection_closed_ok(self):
         """send raises ConnectionClosedOK after a normal closure."""
         await self.remote_connection.close()
@@ -736,14 +793,12 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         # Remove socket.timeout when dropping Python < 3.10.
         self.assertIsInstance(exc.__cause__, (socket.timeout, TimeoutError))
 
-    async def test_close_does_not_wait_for_recv(self):
-        # The asyncio implementation has a buffer for incoming messages. Closing
-        # the connection discards buffered messages. This is allowed by the RFC:
-        # > However, there is no guarantee that the endpoint that has already
-        # > sent a Close frame will continue to process data.
+    async def test_close_preserves_queued_messages(self):
+        """close preserves messages buffered in the assembler."""
         await self.remote_connection.send("😀")
         await self.connection.close()
 
+        self.assertEqual(await self.connection.recv(), "😀")
         with self.assertRaises(ConnectionClosedOK) as raised:
             await self.connection.recv()
 
@@ -837,6 +892,15 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         async with asyncio_timeout(MS):
             await pong_waiter
 
+    async def test_acknowledge_canceled_ping(self):
+        """ping is acknowledged by a pong with the same payload after being canceled."""
+        async with self.drop_frames_rcvd():  # drop automatic response to ping
+            pong_waiter = await self.connection.ping("this")
+        pong_waiter.cancel()
+        await self.remote_connection.pong("this")
+        with self.assertRaises(asyncio.CancelledError):
+            await pong_waiter
+
     async def test_acknowledge_ping_non_matching_pong(self):
         """ping isn't acknowledged by a pong with a different payload."""
         async with self.drop_frames_rcvd():  # drop automatic response to ping
@@ -847,7 +911,7 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
                 await pong_waiter
 
     async def test_acknowledge_previous_ping(self):
-        """ping is acknowledged by a pong with the same payload as a later ping."""
+        """ping is acknowledged by a pong for a later ping."""
         async with self.drop_frames_rcvd():  # drop automatic response to ping
             pong_waiter = await self.connection.ping("this")
             await self.connection.ping("that")
@@ -855,6 +919,18 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         async with asyncio_timeout(MS):
             await pong_waiter
 
+    async def test_acknowledge_previous_canceled_ping(self):
+        """ping is acknowledged by a pong for a later ping after being canceled."""
+        async with self.drop_frames_rcvd():  # drop automatic response to ping
+            pong_waiter = await self.connection.ping("this")
+            pong_waiter_2 = await self.connection.ping("that")
+        pong_waiter.cancel()
+        await self.remote_connection.pong("that")
+        async with asyncio_timeout(MS):
+            await pong_waiter_2
+        with self.assertRaises(asyncio.CancelledError):
+            await pong_waiter
+
     async def test_ping_duplicate_payload(self):
         """ping rejects the same payload until receiving the pong."""
         async with self.drop_frames_rcvd():  # drop automatic response to ping
@@ -1011,15 +1087,26 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         self.assertEqual(connection.close_timeout, 42 * MS)
 
     async def test_max_queue(self):
-        """max_queue parameter configures high-water mark of frames buffer."""
+        """max_queue configures high-water mark of frames buffer."""
         connection = Connection(Protocol(self.LOCAL), max_queue=4)
         transport = Mock()
         connection.connection_made(transport)
         self.assertEqual(connection.recv_messages.high, 4)
 
+    async def test_max_queue_none(self):
+        """max_queue disables high-water mark of frames buffer."""
+        connection = Connection(Protocol(self.LOCAL), max_queue=None)
+        transport = Mock()
+        connection.connection_made(transport)
+        self.assertEqual(connection.recv_messages.high, None)
+        self.assertEqual(connection.recv_messages.low, None)
+
     async def test_max_queue_tuple(self):
-        """max_queue parameter configures high-water mark of frames buffer."""
-        connection = Connection(Protocol(self.LOCAL), max_queue=(4, 2))
+        """max_queue configures high-water and low-water marks of frames buffer."""
+        connection = Connection(
+            Protocol(self.LOCAL),
+            max_queue=(4, 2),
+        )
         transport = Mock()
         connection.connection_made(transport)
         self.assertEqual(connection.recv_messages.high, 4)
@@ -1027,14 +1114,20 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
 
     async def test_write_limit(self):
         """write_limit parameter configures high-water mark of write buffer."""
-        connection = Connection(Protocol(self.LOCAL), write_limit=4096)
+        connection = Connection(
+            Protocol(self.LOCAL),
+            write_limit=4096,
+        )
         transport = Mock()
         connection.connection_made(transport)
         transport.set_write_buffer_limits.assert_called_once_with(4096, None)
 
     async def test_write_limits(self):
         """write_limit parameter configures high and low-water marks of write buffer."""
-        connection = Connection(Protocol(self.LOCAL), write_limit=(4096, 2048))
+        connection = Connection(
+            Protocol(self.LOCAL),
+            write_limit=(4096, 2048),
+        )
         transport = Mock()
         connection.connection_made(transport)
         transport.set_write_buffer_limits.assert_called_once_with(4096, 2048)
@@ -1067,7 +1160,7 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
 
     async def test_state(self):
         """Connection has a state attribute."""
-        self.assertEqual(self.connection.state, State.OPEN)
+        self.assertIs(self.connection.state, State.OPEN)
 
     async def test_request(self):
         """Connection has a request attribute."""
@@ -1081,6 +1174,14 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         """Connection has a subprotocol attribute."""
         self.assertIsNone(self.connection.subprotocol)
 
+    async def test_close_code(self):
+        """Connection has a close_code attribute."""
+        self.assertIsNone(self.connection.close_code)
+
+    async def test_close_reason(self):
+        """Connection has a close_reason attribute."""
+        self.assertIsNone(self.connection.close_reason)
+
     # Test reporting of network errors.
 
     async def test_writing_in_data_received_fails(self):
@@ -1194,7 +1295,7 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
         await self.connection.close()
         await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))
 
-        with self.assertNoLogs():
+        with self.assertNoLogs("websockets", logging.WARNING):
             broadcast([self.connection], "😀")
         await self.assertNoFrameSent()
 
@@ -1205,7 +1306,7 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
             await asyncio.sleep(0)
             await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))
 
-            with self.assertNoLogs():
+            with self.assertNoLogs("websockets", logging.WARNING):
                 broadcast([self.connection], "😀")
             await self.assertNoFrameSent()
 
@@ -1271,7 +1372,10 @@ class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
 
         self.assertEqual(
             [record.getMessage() for record in logs.records],
-            ["skipped broadcast: failed to write message"],
+            [
+                "skipped broadcast: failed to write message: "
+                "RuntimeError: Cannot call write() after write_eof()"
+            ],
         )
 
     @unittest.skipIf(
