@@ -21,6 +21,7 @@ from ..utils import (
     CLIENT_CONTEXT,
     MS,
     SERVER_CONTEXT,
+    AssertNoLogsMixin,
     temp_unix_socket_path,
 )
 from .server import (
@@ -32,7 +33,7 @@ from .server import (
 )
 
 
-class ServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
+class ServerTests(EvalShellMixin, AssertNoLogsMixin, unittest.IsolatedAsyncioTestCase):
     async def test_connection(self):
         """Server receives connection from client and the handshake succeeds."""
         async with serve(*args) as server:
@@ -58,8 +59,7 @@ class ServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
                     await client.recv()
                 self.assertEqual(
                     str(raised.exception),
-                    "received 1011 (internal error); "
-                    "then sent 1011 (internal error)",
+                    "received 1011 (internal error); then sent 1011 (internal error)",
                 )
 
     async def test_existing_socket(self):
@@ -148,14 +148,17 @@ class ServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
         async def handler(ws):
             self.fail("handler must not run")
 
-        async with serve(handler, *args[1:], process_request=process_request) as server:
-            with self.assertRaises(InvalidStatus) as raised:
-                async with connect(get_uri(server)):
-                    self.fail("did not raise")
-            self.assertEqual(
-                str(raised.exception),
-                "server rejected WebSocket connection: HTTP 403",
-            )
+        with self.assertNoLogs("websockets", logging.ERROR):
+            async with serve(
+                handler, *args[1:], process_request=process_request
+            ) as server:
+                with self.assertRaises(InvalidStatus) as raised:
+                    async with connect(get_uri(server)):
+                        self.fail("did not raise")
+                self.assertEqual(
+                    str(raised.exception),
+                    "server rejected WebSocket connection: HTTP 403",
+                )
 
     async def test_async_process_request_returns_response(self):
         """Server aborts handshake if async process_request returns a response."""
@@ -166,44 +169,65 @@ class ServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
         async def handler(ws):
             self.fail("handler must not run")
 
-        async with serve(handler, *args[1:], process_request=process_request) as server:
-            with self.assertRaises(InvalidStatus) as raised:
-                async with connect(get_uri(server)):
-                    self.fail("did not raise")
-            self.assertEqual(
-                str(raised.exception),
-                "server rejected WebSocket connection: HTTP 403",
-            )
+        with self.assertNoLogs("websockets", logging.ERROR):
+            async with serve(
+                handler, *args[1:], process_request=process_request
+            ) as server:
+                with self.assertRaises(InvalidStatus) as raised:
+                    async with connect(get_uri(server)):
+                        self.fail("did not raise")
+                self.assertEqual(
+                    str(raised.exception),
+                    "server rejected WebSocket connection: HTTP 403",
+                )
 
     async def test_process_request_raises_exception(self):
         """Server returns an error if process_request raises an exception."""
 
         def process_request(ws, request):
-            raise RuntimeError
+            raise RuntimeError("BOOM")
 
-        async with serve(*args, process_request=process_request) as server:
-            with self.assertRaises(InvalidStatus) as raised:
-                async with connect(get_uri(server)):
-                    self.fail("did not raise")
-            self.assertEqual(
-                str(raised.exception),
-                "server rejected WebSocket connection: HTTP 500",
-            )
+        with self.assertLogs("websockets", logging.ERROR) as logs:
+            async with serve(*args, process_request=process_request) as server:
+                with self.assertRaises(InvalidStatus) as raised:
+                    async with connect(get_uri(server)):
+                        self.fail("did not raise")
+                self.assertEqual(
+                    str(raised.exception),
+                    "server rejected WebSocket connection: HTTP 500",
+                )
+        self.assertEqual(
+            [record.getMessage() for record in logs.records],
+            ["opening handshake failed"],
+        )
+        self.assertEqual(
+            [str(record.exc_info[1]) for record in logs.records],
+            ["BOOM"],
+        )
 
     async def test_async_process_request_raises_exception(self):
         """Server returns an error if async process_request raises an exception."""
 
         async def process_request(ws, request):
-            raise RuntimeError
+            raise RuntimeError("BOOM")
 
-        async with serve(*args, process_request=process_request) as server:
-            with self.assertRaises(InvalidStatus) as raised:
-                async with connect(get_uri(server)):
-                    self.fail("did not raise")
-            self.assertEqual(
-                str(raised.exception),
-                "server rejected WebSocket connection: HTTP 500",
-            )
+        with self.assertLogs("websockets", logging.ERROR) as logs:
+            async with serve(*args, process_request=process_request) as server:
+                with self.assertRaises(InvalidStatus) as raised:
+                    async with connect(get_uri(server)):
+                        self.fail("did not raise")
+                self.assertEqual(
+                    str(raised.exception),
+                    "server rejected WebSocket connection: HTTP 500",
+                )
+        self.assertEqual(
+            [record.getMessage() for record in logs.records],
+            ["opening handshake failed"],
+        )
+        self.assertEqual(
+            [str(record.exc_info[1]) for record in logs.records],
+            ["BOOM"],
+        )
 
     async def test_process_response_returns_none(self):
         """Server runs process_response but keeps the handshake response."""
@@ -277,31 +301,49 @@ class ServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
         """Server returns an error if process_response raises an exception."""
 
         def process_response(ws, request, response):
-            raise RuntimeError
+            raise RuntimeError("BOOM")
 
-        async with serve(*args, process_response=process_response) as server:
-            with self.assertRaises(InvalidStatus) as raised:
-                async with connect(get_uri(server)):
-                    self.fail("did not raise")
-            self.assertEqual(
-                str(raised.exception),
-                "server rejected WebSocket connection: HTTP 500",
-            )
+        with self.assertLogs("websockets", logging.ERROR) as logs:
+            async with serve(*args, process_response=process_response) as server:
+                with self.assertRaises(InvalidStatus) as raised:
+                    async with connect(get_uri(server)):
+                        self.fail("did not raise")
+                self.assertEqual(
+                    str(raised.exception),
+                    "server rejected WebSocket connection: HTTP 500",
+                )
+        self.assertEqual(
+            [record.getMessage() for record in logs.records],
+            ["opening handshake failed"],
+        )
+        self.assertEqual(
+            [str(record.exc_info[1]) for record in logs.records],
+            ["BOOM"],
+        )
 
     async def test_async_process_response_raises_exception(self):
         """Server returns an error if async process_response raises an exception."""
 
         async def process_response(ws, request, response):
-            raise RuntimeError
+            raise RuntimeError("BOOM")
 
-        async with serve(*args, process_response=process_response) as server:
-            with self.assertRaises(InvalidStatus) as raised:
-                async with connect(get_uri(server)):
-                    self.fail("did not raise")
-            self.assertEqual(
-                str(raised.exception),
-                "server rejected WebSocket connection: HTTP 500",
-            )
+        with self.assertLogs("websockets", logging.ERROR) as logs:
+            async with serve(*args, process_response=process_response) as server:
+                with self.assertRaises(InvalidStatus) as raised:
+                    async with connect(get_uri(server)):
+                        self.fail("did not raise")
+                self.assertEqual(
+                    str(raised.exception),
+                    "server rejected WebSocket connection: HTTP 500",
+                )
+        self.assertEqual(
+            [record.getMessage() for record in logs.records],
+            ["opening handshake failed"],
+        )
+        self.assertEqual(
+            [str(record.exc_info[1]) for record in logs.records],
+            ["BOOM"],
+        )
 
     async def test_override_server(self):
         """Server can override Server header with server_header."""
@@ -430,6 +472,10 @@ class ServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
         )
         self.assertEqual(
             [str(record.exc_info[1]) for record in logs.records],
+            ["did not receive a valid HTTP request"],
+        )
+        self.assertEqual(
+            [str(record.exc_info[1].__cause__) for record in logs.records],
             ["invalid HTTP request line: HELO relay.invalid"],
         )
 
@@ -717,7 +763,7 @@ class BasicAuthTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
 
     async def test_without_credentials_or_check_credentials(self):
         """basic_auth requires either credentials or check_credentials."""
-        with self.assertRaises(TypeError) as raised:
+        with self.assertRaises(ValueError) as raised:
             basic_auth()
         self.assertEqual(
             str(raised.exception),
@@ -726,7 +772,7 @@ class BasicAuthTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
 
     async def test_with_credentials_and_check_credentials(self):
         """basic_auth requires only one of credentials and check_credentials."""
-        with self.assertRaises(TypeError) as raised:
+        with self.assertRaises(ValueError) as raised:
             basic_auth(
                 credentials=("hello", "iloveyou"),
                 check_credentials=lambda: False,  # pragma: no cover
