@@ -1,6 +1,5 @@
 import contextlib
 import logging
-import platform
 import socket
 import sys
 import threading
@@ -15,7 +14,7 @@ from websockets.exceptions import (
     ConnectionClosedOK,
 )
 from websockets.frames import CloseCode, Frame, Opcode
-from websockets.protocol import CLIENT, SERVER, Protocol
+from websockets.protocol import CLIENT, SERVER, Protocol, State
 from websockets.sync.connection import *
 
 from ..protocol import RecordingProtocol
@@ -154,6 +153,16 @@ class ClientConnectionTests(unittest.TestCase):
         self.remote_connection.send(b"\x01\x02\xfe\xff")
         self.assertEqual(self.connection.recv(), b"\x01\x02\xfe\xff")
 
+    def test_recv_text_as_bytes(self):
+        """recv receives a text message as bytes."""
+        self.remote_connection.send("😀")
+        self.assertEqual(self.connection.recv(decode=False), "😀".encode())
+
+    def test_recv_binary_as_text(self):
+        """recv receives a binary message as a str."""
+        self.remote_connection.send("😀".encode())
+        self.assertEqual(self.connection.recv(decode=True), "😀")
+
     def test_recv_fragmented_text(self):
         """recv receives a fragmented text message."""
         self.remote_connection.send(["😀", "😀"])
@@ -176,6 +185,15 @@ class ClientConnectionTests(unittest.TestCase):
         with self.assertRaises(ConnectionClosedError):
             self.connection.recv()
 
+    def test_recv_non_utf8_text(self):
+        """recv receives a non-UTF-8 text message."""
+        self.remote_connection.send(b"\x01\x02\xfe\xff", text=True)
+        with self.assertRaises(ConnectionClosedError):
+            self.connection.recv()
+        self.assertFrameSent(
+            Frame(Opcode.CLOSE, b"\x03\xefinvalid start byte at position 2")
+        )
+
     def test_recv_during_recv(self):
         """recv raises ConcurrencyError when called concurrently."""
         recv_thread = threading.Thread(target=self.connection.recv)
@@ -228,6 +246,22 @@ class ClientConnectionTests(unittest.TestCase):
             [b"\x01\x02\xfe\xff"],
         )
 
+    def test_recv_streaming_text_as_bytes(self):
+        """recv_streaming receives a text message as bytes."""
+        self.remote_connection.send("😀")
+        self.assertEqual(
+            list(self.connection.recv_streaming(decode=False)),
+            ["😀".encode()],
+        )
+
+    def test_recv_streaming_binary_as_str(self):
+        """recv_streaming receives a binary message as a str."""
+        self.remote_connection.send("😀".encode())
+        self.assertEqual(
+            list(self.connection.recv_streaming(decode=True)),
+            ["😀"],
+        )
+
     def test_recv_streaming_fragmented_text(self):
         """recv_streaming receives a fragmented text message."""
         self.remote_connection.send(["😀", "😀"])
@@ -260,6 +294,15 @@ class ClientConnectionTests(unittest.TestCase):
             for _ in self.connection.recv_streaming():
                 self.fail("did not raise")
 
+    def test_recv_streaming_non_utf8_text(self):
+        """recv_streaming receives a non-UTF-8 text message."""
+        self.remote_connection.send(b"\x01\x02\xfe\xff", text=True)
+        with self.assertRaises(ConnectionClosedError):
+            list(self.connection.recv_streaming())
+        self.assertFrameSent(
+            Frame(Opcode.CLOSE, b"\x03\xefinvalid start byte at position 2")
+        )
+
     def test_recv_streaming_during_recv(self):
         """recv_streaming raises ConcurrencyError when called concurrently with recv."""
         recv_thread = threading.Thread(target=self.connection.recv)
@@ -308,6 +351,16 @@ class ClientConnectionTests(unittest.TestCase):
         self.connection.send(b"\x01\x02\xfe\xff")
         self.assertEqual(self.remote_connection.recv(), b"\x01\x02\xfe\xff")
 
+    def test_send_binary_from_str(self):
+        """send sends a binary message from a str."""
+        self.connection.send("😀", text=False)
+        self.assertEqual(self.remote_connection.recv(), "😀".encode())
+
+    def test_send_text_from_bytes(self):
+        """send sends a text message from bytes."""
+        self.connection.send("😀".encode(), text=True)
+        self.assertEqual(self.remote_connection.recv(), "😀")
+
     def test_send_fragmented_text(self):
         """send sends a fragmented text message."""
         self.connection.send(["😀", "😀"])
@@ -326,6 +379,24 @@ class ClientConnectionTests(unittest.TestCase):
             [b"\x01\x02", b"\xfe\xff", b""],
         )
 
+    def test_send_fragmented_binary_from_str(self):
+        """send sends a fragmented binary message from a str."""
+        self.connection.send(["😀", "😀"], text=False)
+        # websockets sends an trailing empty fragment. That's an implementation detail.
+        self.assertEqual(
+            list(self.remote_connection.recv_streaming()),
+            ["😀".encode(), "😀".encode(), b""],
+        )
+
+    def test_send_fragmented_text_from_bytes(self):
+        """send sends a fragmented text message from bytes."""
+        self.connection.send(["😀".encode(), "😀".encode()], text=True)
+        # websockets sends an trailing empty fragment. That's an implementation detail.
+        self.assertEqual(
+            list(self.remote_connection.recv_streaming()),
+            ["😀", "😀", ""],
+        )
+
     def test_send_connection_closed_ok(self):
         """send raises ConnectionClosedOK after a normal closure."""
         self.remote_connection.close()
@@ -471,28 +542,12 @@ class ClientConnectionTests(unittest.TestCase):
         # Remove socket.timeout when dropping Python < 3.10.
         self.assertIsInstance(exc.__cause__, (socket.timeout, TimeoutError))
 
-    def test_close_waits_for_recv(self):
-        # The sync implementation doesn't have a buffer for incoming messsages.
-        # It requires reading incoming frames until the close frame is reached.
-        # This behavior — close() blocks until recv() is called — is less than
-        # ideal and inconsistent with the asyncio implementation.
+    def test_close_preserves_queued_messages(self):
+        """close preserves messages buffered in the assembler."""
         self.remote_connection.send("😀")
+        self.connection.close()
 
-        close_thread = threading.Thread(target=self.connection.close)
-        close_thread.start()
-
-        # Let close() initiate the closing handshake and send a close frame.
-        time.sleep(MS)
-        self.assertTrue(close_thread.is_alive())
-
-        # Connection isn't closed yet.
-        self.connection.recv()
-
-        # Let close() receive a close frame and finish the closing handshake.
-        time.sleep(MS)
-        self.assertFalse(close_thread.is_alive())
-
-        # Connection is closed now.
+        self.assertEqual(self.connection.recv(), "😀")
         with self.assertRaises(ConnectionClosedOK) as raised:
             self.connection.recv()
 
@@ -500,24 +555,6 @@ class ClientConnectionTests(unittest.TestCase):
         self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
         self.assertIsNone(exc.__cause__)
 
-    def test_close_timeout_waiting_for_recv(self):
-        self.remote_connection.send("😀")
-
-        close_thread = threading.Thread(target=self.connection.close)
-        close_thread.start()
-
-        # Let close() time out during the closing handshake.
-        time.sleep(3 * MS)
-        self.assertFalse(close_thread.is_alive())
-
-        # Connection is closed now.
-        with self.assertRaises(ConnectionClosedError) as raised:
-            self.connection.recv()
-
-        exc = raised.exception
-        self.assertEqual(str(exc), "sent 1000 (OK); no close frame received")
-        self.assertIsInstance(exc.__cause__, TimeoutError)
-
     def test_close_idempotency(self):
         """close does nothing if the connection is already closed."""
         self.connection.close()
@@ -526,17 +563,13 @@ class ClientConnectionTests(unittest.TestCase):
         self.connection.close()
         self.assertNoFrameSent()
 
-    @unittest.skipIf(
-        platform.python_implementation() == "PyPy",
-        "this test fails randomly due to a bug in PyPy",  # see #1314 for details
-    )
     def test_close_idempotency_race_condition(self):
         """close waits if the connection is already closing."""
 
-        self.connection.close_timeout = 5 * MS
+        self.connection.close_timeout = 6 * MS
 
         def closer():
-            with self.delay_frames_rcvd(3 * MS):
+            with self.delay_frames_rcvd(4 * MS):
                 self.connection.close()
 
         close_thread = threading.Thread(target=closer)
@@ -548,14 +581,14 @@ class ClientConnectionTests(unittest.TestCase):
 
         # Connection isn't closed yet.
         with self.assertRaises(TimeoutError):
-            self.connection.recv(timeout=0)
+            self.connection.recv(timeout=MS)
 
         self.connection.close()
         self.assertNoFrameSent()
 
         # Connection is closed now.
         with self.assertRaises(ConnectionClosedOK):
-            self.connection.recv(timeout=0)
+            self.connection.recv(timeout=MS)
 
         close_thread.join()
 
@@ -645,7 +678,7 @@ class ClientConnectionTests(unittest.TestCase):
         self.assertFalse(pong_waiter.wait(MS))
 
     def test_acknowledge_previous_ping(self):
-        """ping is acknowledged by a pong with the same payload as a later ping."""
+        """ping is acknowledged by a pong for as a later ping."""
         with self.drop_frames_rcvd():  # drop automatic response to ping
             pong_waiter = self.connection.ping("this")
             self.connection.ping("that")
@@ -696,6 +729,58 @@ class ClientConnectionTests(unittest.TestCase):
         with self.assertRaises(TypeError):
             self.connection.pong([])
 
+    # Test parameters.
+
+    def test_close_timeout(self):
+        """close_timeout parameter configures close timeout."""
+        socket_, remote_socket = socket.socketpair()
+        self.addCleanup(socket_.close)
+        self.addCleanup(remote_socket.close)
+        connection = Connection(
+            socket_,
+            Protocol(self.LOCAL),
+            close_timeout=42 * MS,
+        )
+        self.assertEqual(connection.close_timeout, 42 * MS)
+
+    def test_max_queue(self):
+        """max_queue configures high-water mark of frames buffer."""
+        socket_, remote_socket = socket.socketpair()
+        self.addCleanup(socket_.close)
+        self.addCleanup(remote_socket.close)
+        connection = Connection(
+            socket_,
+            Protocol(self.LOCAL),
+            max_queue=4,
+        )
+        self.assertEqual(connection.recv_messages.high, 4)
+
+    def test_max_queue_none(self):
+        """max_queue disables high-water mark of frames buffer."""
+        socket_, remote_socket = socket.socketpair()
+        self.addCleanup(socket_.close)
+        self.addCleanup(remote_socket.close)
+        connection = Connection(
+            socket_,
+            Protocol(self.LOCAL),
+            max_queue=None,
+        )
+        self.assertEqual(connection.recv_messages.high, None)
+        self.assertEqual(connection.recv_messages.high, None)
+
+    def test_max_queue_tuple(self):
+        """max_queue configures high-water and low-water marks of frames buffer."""
+        socket_, remote_socket = socket.socketpair()
+        self.addCleanup(socket_.close)
+        self.addCleanup(remote_socket.close)
+        connection = Connection(
+            socket_,
+            Protocol(self.LOCAL),
+            max_queue=(4, 2),
+        )
+        self.assertEqual(connection.recv_messages.high, 4)
+        self.assertEqual(connection.recv_messages.low, 2)
+
     # Test attributes.
 
     def test_id(self):
@@ -718,6 +803,10 @@ class ClientConnectionTests(unittest.TestCase):
         self.assertEqual(self.connection.remote_address, ("peer", 1234))
         getpeername.assert_called_with()
 
+    def test_state(self):
+        """Connection has a state attribute."""
+        self.assertIs(self.connection.state, State.OPEN)
+
     def test_request(self):
         """Connection has a request attribute."""
         self.assertIsNone(self.connection.request)
@@ -730,6 +819,14 @@ class ClientConnectionTests(unittest.TestCase):
         """Connection has a subprotocol attribute."""
         self.assertIsNone(self.connection.subprotocol)
 
+    def test_close_code(self):
+        """Connection has a close_code attribute."""
+        self.assertIsNone(self.connection.close_code)
+
+    def test_close_reason(self):
+        """Connection has a close_reason attribute."""
+        self.assertIsNone(self.connection.close_reason)
+
     # Test reporting of network errors.
 
     @unittest.skipUnless(sys.platform == "darwin", "works only on BSD")
